import sqlite3
import numpy as np
import pandas as pd
from datetime import datetime
from info_ratio.common.input_output import get_data


def preprocess_data(self):

    start = pd.Timestamp(self.start_date)
    end = pd.Timestamp(self.end_date)

    # Connect to SQLite database
    with sqlite3.connect(self.db_path) as db:
        db = get_data(db, insert_data=self.insert_data)

        # Fetch data from tabless
        fund_data_df = pd.read_sql_query("SELECT * FROM df_table", db)
        company_data_df = pd.read_sql_query("SELECT * FROM company_table", db)

    fund_data_df['Numeric Date'] = fund_data_df['Port Date'].apply(
        lambda x: datetime.strptime(x, "%d-%b-%Y").strftime("%Y%m")).astype(np.int64)

    # Merge single_fund_df with ace_df_renamed on 'Company ISIN' and 'Numeric Date'
    # Use 'left' join to keep all rows from single_fund_df and fill in 'Price' where matches are found
    merged_df = pd.merge(fund_data_df, company_data_df, on=[
                         'Company ISIN', 'Numeric Date'], how='left')
    merged_df["No Of Shares"] = merged_df["No Of Shares"].replace(
        '--', None, regex=True)

    # Drop rows where 'Price' and 'No of Shares'is missing
    merged_df = merged_df[merged_df['Price'].notna()]
    merged_df = merged_df[merged_df["No Of Shares"].notna()]

    # Convert 'Port Date' to a datetime object
    merged_df['Port Date_'] = pd.to_datetime(merged_df['Port Date'])
    merged_df = merged_df.sort_values(by='Port Date_')

    merged_df["No Of Shares"] = merged_df["No Of Shares"].astype(int)
    merged_df = merged_df.loc[merged_df['Port Date_'].between(start, end)]

    return merged_df


def calculate_information_ratio(input_df, fund_info):

    num_wins = 0
    num_losses = 0
    absolute_gain = 0
    absolute_loss = 0

    # Get the breadth of the fund
    stock_list = input_df['Company Name'].unique()
    breadth = len(stock_list)
    if breadth == 0:
        return fund_info, 4
    fund_info.append(breadth)

    for i in stock_list:
        stock_df = input_df[input_df['Company Name'] == i]
        stock_df["diff_tmp"] = stock_df["No Of Shares"].diff()
        stock_df["ret_tmp"] = stock_df["Price"] * stock_df["diff_tmp"]
        stock_df['ret_tmp'].iloc[0] = stock_df['Price'].iloc[0] * \
            stock_df['No Of Shares'].iloc[0]

        # We are assuming the the portfolio manager has exited the remaining shares at the end of the analysis period
        last_exit = stock_df['Price'].iloc[-1] * \
            stock_df['No Of Shares'].iloc[-1]

        sum_of_positive_nos = stock_df[stock_df['ret_tmp']
                                       >= 0]['ret_tmp'].sum()

        some_of_negative_nos = np.abs(stock_df[stock_df['ret_tmp'] <= 0]
                                      ['ret_tmp']).sum() + last_exit

        if sum_of_positive_nos < some_of_negative_nos:
            num_wins += 1
            absolute_gain += round((some_of_negative_nos -
                                   sum_of_positive_nos), 3)
        else:
            num_losses += 1
            absolute_loss += round((some_of_negative_nos -
                                   sum_of_positive_nos), 3)

    if num_wins == 0:
        return fund_info, 0
    else:
        average_gain = abs(absolute_gain)/num_wins
    if num_losses == 0:
        return fund_info, 1
    else:
        average_loss = abs(absolute_loss)/num_losses

    fund_info.append(num_wins)
    fund_info.append(num_losses)

    # Pain-to-Gain Ratio
    pain_gain_ratio = abs(absolute_gain/absolute_loss)

    # Calculate the batting average
    batting_average = num_wins/breadth
    fund_info.append(round(batting_average, 3))

    # Calculate the slugging ratio
    slugging_ratio = average_gain/average_loss

    fund_info.append(round(slugging_ratio, 3))

    information_coefficient = 1.6*(batting_average - (1/(1+slugging_ratio)))
    fund_info.append(round(information_coefficient, 3))

    information_ratio = information_coefficient * (np.sqrt(breadth))
    fund_info.append(round(information_ratio, 3))
    fund_info.append(round(pain_gain_ratio, 3))

    return fund_info, 3



import pandas as pd


def dfs_tabs(df_list, sheet_list, file_name):
    writer = pd.ExcelWriter(file_name, engine='openpyxl')
    for dataframe, sheet in zip(df_list, sheet_list):
        dataframe.to_excel(writer, sheet_name=sheet,
                           startrow=0, startcol=0, index=False)
    writer.close()


def get_data(db, insert_data=False):
    if insert_data:
        # Define the path to the Excel file
        file_uploads = ["info_ratio/data/Holdings Historical 360 to DSP.xlsx",
                        "info_ratio/data/Holdings Historical Edelweiss to Invesco.xlsx",
                        "info_ratio/data/Holdings Historical ITI to PPFAS.xlsx",
                        "info_ratio/data/Holdings Historical Quant to Zerodha.xlsx"]

        # Loop through each Excel file in 'file_uploads'
        for i in file_uploads:
            dfs = pd.read_excel(i, engine="openpyxl", skiprows=4)
            dfs.to_sql('df_table', con=db, if_exists='append')

        cols = ['CD_ISIN No', 'Bmonth_Month End', 'Bmonth_Close']
        col_rename = {'CD_ISIN No': 'Company ISIN',
                      'Bmonth_Month End': 'Numeric Date', 'Bmonth_Close': 'Price'}
        company_data_df = pd.read_excel(
            "info_ratio/data/Company_Data.xlsx", engine="openpyxl", usecols=cols)
        company_data_dump = company_data_df.rename(columns=col_rename)
        company_data_dump.to_sql('company_table', con=db, if_exists='append')

        return db
    else:
        return db



from loga import Loga

# all setup values are optional
loga = Loga(
    facility="info_ratio",  # name of program logging the message
    do_print=True,  # print each log to console
    truncation=1000,  # longest possible value in extra data
    private_data={"password"},  # set of sensitive args/kwargs
)



import os
from info_ratio.common.logging import loga
import pandas as pd
from datetime import datetime
from info_ratio.common.input_output import dfs_tabs
from info_ratio.common.helper_functions import preprocess_data, calculate_information_ratio


pd.options.mode.chained_assignment = None  # default='warn'


@loga
class InfoRatio:
    def __init__(self, funds_for_analysis_path, output_path, db_path, start_date, end_date, insert_data=False, enable_logging=True):

        if not enable_logging:
            loga.stop()

        self.main_dict = []
        self.only_wins = []
        self.only_losses = []
        self.dont_exist = []

        self.insert_data = insert_data

        self.start_date = start_date
        self.end_date = end_date

        self.funds_for_analysis_path = funds_for_analysis_path
        self.output_path = output_path
        self.db_path = db_path

    def run(self):
        input_df = preprocess_data(self)

        funds_df = pd.read_excel(
            self.funds_for_analysis_path, sheet_name='Enter Funds', engine="openpyxl")

        for scheme_name in funds_df['Funds']:
            fund_info = [scheme_name]
            scheme_df = input_df[input_df['Scheme Name'] == scheme_name]

            fund_info, type_ = calculate_information_ratio(
                scheme_df, fund_info)
            if type_ == 0:
                self.only_losses.append(fund_info)
            elif type_ == 1:
                self.only_wins.append(fund_info)
            elif type_ == 4:
                self.dont_exist.append(fund_info)
            else:
                self.main_dict.append(fund_info)

        df = pd.DataFrame(self.main_dict, columns=['Fund Name', 'Breadth', 'Wins', 'Loss', "Batting Average",
                          "Slugging Ratio", "Information Coefficient", "Information Ratio", "Pain/Gain Ratio"])
        only_win_df = pd.DataFrame(self.only_wins, columns=[
                                   'Fund Name', 'Breadth'])
        only_loss_df = pd.DataFrame(self.only_losses, columns=[
                                    'Fund Name', 'Breadth'])
        dont_exist_df = pd.DataFrame(self.dont_exist, columns=['Fund Name'])

        # list of dataframes and sheet names
        dfs = [df, only_win_df, only_loss_df, dont_exist_df]
        sheets = ['Main', 'Only Wins', 'Only Losses', "Don't Exist"]

        name = self.output_path + datetime.now().strftime("%Y-%m-%d")+".xlsx"
        # run function
        dfs_tabs(dfs, sheets, name)
